from ..enums import CalculationRate
from .core import OutputProxy, UGen, param, ugen


@ugen(ar=True, kr=True, ir=True)
class Clip(UGen):
    """
    Clips a signal outside given thresholds.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> clip = supriya.ugens.Clip.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> clip
        <Clip.ar()[0]>
    """

    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True, ir=True)
class Fold(UGen):
    """
    Folds a signal outside given thresholds.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> fold = supriya.ugens.Fold.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> fold
        <Fold.ar()[0]>
    """

    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True)
class Gate(UGen):
    """
    Gates or holds.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> trigger = supriya.ugens.Dust.kr(density=1)
        >>> gate = supriya.ugens.Gate.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> gate
        <Gate.ar()[0]>
    """

    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True, ir=True)
class InRange(UGen):
    """
    Tests if a signal is within a given range.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> in_range = supriya.ugens.InRange.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> in_range
        <InRange.ar()[0]>
    """

    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True)
class Latch(UGen):
    """
    Samples and holds.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> trigger = supriya.ugens.Dust.kr(density=1)
        >>> latch = supriya.ugens.Latch.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> latch
        <Latch.ar()[0]>
    """

    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class LeastChange(UGen):
    """
    Outputs least changed input.

    ::

        >>> least_change = supriya.ugens.LeastChange.ar(
        ...     a=0,
        ...     b=0,
        ... )
        >>> least_change
        <LeastChange.ar()[0]>
    """

    a = param(0)
    b = param(0)


@ugen(ar=True, kr=True)
class MostChange(UGen):
    """
    Outputs most changed input.

    ::

        >>> most_change = supriya.ugens.MostChange.ar(
        ...     a=0,
        ...     b=0,
        ... )
        >>> most_change
        <MostChange.ar()[0]>
    """

    a = param(0)
    b = param(0)


@ugen(ar=True, kr=True)
class Peak(UGen):
    """
    Tracks peak signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
        >>> peak = supriya.ugens.Peak.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> peak
        <Peak.ar()[0]>
    """

    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class PeakFollower(UGen):
    """
    Tracks peak signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> peak_follower = supriya.ugens.PeakFollower.ar(
        ...     decay=0.999,
        ...     source=source,
        ... )
        >>> peak_follower
        <PeakFollower.ar()[0]>
    """

    source = param()
    decay = param(0.999)


@ugen(ar=True, kr=True)
class Phasor(UGen):
    """
    A resettable linear ramp between two levels.

    ::

        >>> trigger = supriya.ugens.Impulse.kr(frequency=0.5)
        >>> phasor = supriya.ugens.Phasor.ar(
        ...     rate=1,
        ...     reset_pos=0,
        ...     start=0,
        ...     stop=1,
        ...     trigger=trigger,
        ... )
        >>> phasor
        <Phasor.ar()[0]>
    """

    trigger = param(0)
    rate = param(1.0)
    start = param(0.0)
    stop = param(1.0)
    reset_pos = param(0.0)


@ugen(ar=True, kr=True)
class Poll(UGen):
    """
    A UGen poller.

    ::

        >>> sine = supriya.ugens.SinOsc.ar()
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
        >>> poll = supriya.ugens.Poll.ar(
        ...     source=sine,
        ...     trigger=trigger,
        ...     trigger_id=1234,
        ... )
        >>> poll
        <Poll.ar()[0]>

    .. container:: example

        Unlike **sclang**, Python does not share any inter-process communication with
        **scsynth**. This means that the Poll UGen is not able to automatically print
        out its diagnostic messages into a Python interpreter session.

        To get information out of the Poll UGen, we first need to set the Poll's
        `trigger_id` to a value greater than 0. This will cause the poll to send `/tr`
        OSC messages back to its client - Python. We can then register a callback to
        respond to these `/tr` messages.

        ::

            >>> with supriya.SynthDefBuilder() as builder:
            ...     sine = supriya.ugens.SinOsc.ar()
            ...     trigger = supriya.ugens.Impulse.kr(frequency=1)
            ...     poll = supriya.ugens.Poll.ar(
            ...         source=sine,
            ...         trigger=trigger,
            ...         trigger_id=1234,
            ...     )
            ...
            >>> synthdef = builder.build()

        ::

            >>> server = supriya.Server().boot(port=supriya.osc.find_free_port())
            >>> _ = server.add_synthdefs(
            ...     synthdef,
            ...     on_completion=lambda context: context.add_synth(synthdef),
            ... )
            >>> _ = server.sync()
            >>> callback = server.osc_protocol.register(
            ...     pattern="/tr",
            ...     procedure=lambda message: print(
            ...         "Polled: {!r}".format(message)
            ...     ),
            ...     once=True,
            ... )

        ::

            >>> _ = server.quit()
    """

    ### CLASS VARIABLES ###

    trigger = param()
    source = param()
    trigger_id = param(-1)
    label = param(unexpanded=True)

    ### INITIALIZER ###

    def __init__(
        self,
        calculation_rate=None,
        label=None,
        source=None,
        special_index=None,
        trigger=None,
        trigger_id=-1,
    ):
        if label is None:
            if isinstance(source, UGen):
                label = type(source).__name__
            elif isinstance(source, OutputProxy):
                label = type(source.ugen).__name__
        label = str(label)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            label=[len(label), *(ord(c) for c in label)],
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )

    ### PUBLIC METHODS ###

    @classmethod
    def ar(cls, label=None, source=None, trigger=None, trigger_id=-1):
        return cls._new_expanded(
            calculation_rate=CalculationRate.AUDIO,
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )

    @classmethod
    def kr(cls, label=None, source=None, trigger=None, trigger_id=-1):
        return cls._new_expanded(
            calculation_rate=CalculationRate.CONTROL,
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )

    @classmethod
    def new(cls, label=None, source=None, trigger=None, trigger_id=-1):
        return cls._new_expanded(
            calculation_rate=[CalculationRate.from_expr(x) for x in source],
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )


@ugen(ar=True, kr=True)
class RunningMax(UGen):
    """
    Tracks maximum signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
        >>> running_max = supriya.ugens.RunningMax.ar(
        ...     source=source,
        ...     trigger=0,
        ... )
        >>> running_max
        <RunningMax.ar()[0]>
    """

    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class RunningMin(UGen):
    """
    Tracks minimum signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
        >>> running_min = supriya.ugens.RunningMin.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> running_min
        <RunningMin.ar()[0]>
    """

    source = param()
    trigger = param(0)


@ugen(ar=True, kr=True)
class Schmidt(UGen):
    """
    A Schmidt trigger.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> schmidt = supriya.ugens.Schmidt.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> schmidt
        <Schmidt.ar()[0]>
    """

    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True, channel_count=0, fixed_channel_count=True)
class SendPeakRMS(UGen):
    """
    Tracks peak and power of a signal for GUI applications.

    ::

        >>> send_peak_rms = supriya.ugens.SendPeakRMS.kr(
        ...     command_name="/reply",
        ...     peak_lag=3,
        ...     reply_id=-1,
        ...     reply_rate=20,
        ...     source=[1, 2, 3],
        ... )
        >>> send_peak_rms
        <SendPeakRMS.kr()>
    """

    ### CLASS VARIABLES ###

    reply_rate = param(20)
    peak_lag = param(3)
    reply_id = param(-1)
    source_size = param()
    source = param(unexpanded=True)
    character_count = param()
    character = param(unexpanded=True)

    ### PUBLIC METHODS ###

    @classmethod
    def ar(
        cls, command_name="/reply", peak_lag=3, reply_id=-1, reply_rate=20, source=None
    ):
        """
        Constructs an audio-rate SendPeakRMS.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> send_peak_rms = supriya.ugens.SendPeakRMS.ar(
            ...     command_name="/reply",
            ...     peak_lag=3,
            ...     reply_id=-1,
            ...     reply_rate=20,
            ...     source=source,
            ... )
            >>> send_peak_rms
            <SendPeakRMS.ar()>

        Returns ugen graph.
        """
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.AUDIO,
            peak_lag=peak_lag,
            reply_id=reply_id,
            reply_rate=reply_rate,
            source=source,
            source_size=len(source),
            character_count=len(command),
            character=[ord(x) for x in command],
        )

    @classmethod
    def kr(
        cls, command_name="/reply", peak_lag=3, reply_id=-1, reply_rate=20, source=None
    ):
        """
        Constructs a control-rate SendPeakRMS.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> send_peak_rms = supriya.ugens.SendPeakRMS.kr(
            ...     command_name="/reply",
            ...     peak_lag=3,
            ...     reply_id=-1,
            ...     reply_rate=20,
            ...     source=source,
            ... )
            >>> send_peak_rms
            <SendPeakRMS.kr()>

        Returns ugen graph.
        """
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.CONTROL,
            peak_lag=peak_lag,
            reply_id=reply_id,
            reply_rate=reply_rate,
            source=source,
            source_size=len(source),
            character_count=len(command),
            character=[ord(x) for x in command],
        )


@ugen(ar=True, kr=True, channel_count=0, fixed_channel_count=True)
class SendReply(UGen):
    """
    Sends an array of values from the server to all notified clients.

    ::

        >>> source = supriya.ugens.In.ar(channel_count=4)
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
        >>> send_reply = supriya.ugens.SendReply.kr(
        ...     command_name="/reply",
        ...     source=source,
        ...     trigger=trigger,
        ... )
    """

    trigger = param()
    reply_id = param(-1)
    character_count = param()
    character = param(unexpanded=True)
    source = param(unexpanded=True)

    @classmethod
    def ar(cls, command_name="/reply", reply_id=-1, source=None, trigger=None):
        """
        Constructs an audio-rate SendReply.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
            >>> send_reply = supriya.ugens.SendReply.ar(
            ...     command_name="/reply",
            ...     source=source,
            ...     trigger=trigger,
            ... )
            >>> send_reply
            <SendReply.ar()>

        Returns ugen graph.
        """
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.AUDIO,
            trigger=trigger,
            reply_id=reply_id,
            character_count=len(command),
            character=[ord(x) for x in command],
            source=source,
        )

    @classmethod
    def kr(cls, command_name="/reply", reply_id=-1, source=None, trigger=None):
        """
        Constructs a control-rate SendReply.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> trigger = supriya.ugens.Impulse.kr(frequency=1)
            >>> send_reply = supriya.ugens.SendReply.kr(
            ...     command_name="/reply",
            ...     source=source,
            ...     trigger=trigger,
            ... )
            >>> send_reply
            <SendReply.kr()>

        Returns ugen graph.
        """
        command = str(command_name)
        return cls._new_single(
            calculation_rate=CalculationRate.CONTROL,
            trigger=trigger,
            reply_id=reply_id,
            character_count=len(command),
            character=[ord(x) for x in command],
            source=source,
        )


@ugen(ar=True, kr=True)
class SendTrig(UGen):
    trigger = param()
    id_ = param(0)
    value = param(0.0)


@ugen(ar=True, kr=True)
class Sweep(UGen):
    """
    A triggered linear ramp.

    ::

        >>> sweep = supriya.ugens.Sweep.ar(
        ...     rate=1,
        ...     trigger=0,
        ... )
        >>> sweep
        <Sweep.ar()[0]>
    """

    trigger = param(0)
    rate = param(1.0)


@ugen(ar=True, kr=True)
class TDelay(UGen):
    """
    A trigger delay.

    ::

        >>> source = supriya.ugens.Dust.kr(density=1)
        >>> tdelay = supriya.ugens.TDelay.ar(
        ...     duration=0.1,
        ...     source=source,
        ... )
        >>> tdelay
        <TDelay.ar()[0]>
    """

    source = param()
    duration = param(0.1)


@ugen(ar=True, kr=True)
class ToggleFF(UGen):
    """
    A toggle flip-flop.

    ::

        >>> trigger = supriya.ugens.Dust.kr(density=1)
        >>> toggle_ff = supriya.ugens.ToggleFF.ar(
        ...     trigger=trigger,
        ... )
        >>> toggle_ff
        <ToggleFF.ar()[0]>
    """

    trigger = param(0)


@ugen(ar=True, kr=True)
class Trig1(UGen):
    """
    A timed trigger.

    ::

        >>> source = supriya.ugens.Dust.kr(density=1)
        >>> trig_1 = supriya.ugens.Trig1.ar(
        ...     duration=0.1,
        ...     source=source,
        ... )
        >>> trig_1
        <Trig1.ar()[0]>
    """

    source = param()
    duration = param(0.1)


@ugen(ar=True, kr=True)
class Trig(UGen):
    """
    A timed trigger.

    ::

        >>> source = supriya.ugens.Dust.kr(density=1)
        >>> trig = supriya.ugens.Trig.ar(
        ...     duration=0.1,
        ...     source=source,
        ... )
        >>> trig
        <Trig.ar()[0]>
    """

    source = param()
    duration = param(0.1)


@ugen(ar=True, kr=True, ir=True)
class Wrap(UGen):
    """
    Wraps a signal outside given thresholds.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> wrap = supriya.ugens.Wrap.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> wrap
        <Wrap.ar()[0]>
    """

    source = param()
    minimum = param(0.0)
    maximum = param(1.0)


@ugen(ar=True, kr=True)
class ZeroCrossing(UGen):
    """
    A zero-crossing frequency follower.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> zero_crossing = supriya.ugens.ZeroCrossing.ar(
        ...     source=source,
        ... )
        >>> zero_crossing
        <ZeroCrossing.ar()[0]>
    """

    source = param()



from .core import UGen, param, ugen


@ugen(ar=True)
class CuspL(UGen):
    """
    A linear-interpolating cusp map chaotic generator.

    ::

        >>> cusp_l = supriya.ugens.CuspL.ar(
        ...     a=1,
        ...     b=1.9,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> cusp_l
        <CuspL.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.0)
    b = param(1.9)
    xi = param(0.0)


@ugen(ar=True)
class CuspN(UGen):
    """
    A non-interpolating cusp map chaotic generator.

    ::

        >>> cusp_n = supriya.ugens.CuspN.ar(
        ...     a=1,
        ...     b=1.9,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> cusp_n
        <CuspN.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.0)
    b = param(1.9)
    xi = param(0.0)


@ugen(ar=True)
class FBSineC(UGen):
    """
    A cubic-interpolating feedback sine with chaotic phase indexing.

    ::

        >>> fbsine_c = supriya.ugens.FBSineC.ar(
        ...     a=1.1,
        ...     c=0.5,
        ...     fb=0.1,
        ...     frequency=22050,
        ...     im=1,
        ...     xi=0.1,
        ...     yi=0.1,
        ... )
        >>> fbsine_c
        <FBSineC.ar()[0]>
    """

    frequency = param(22050)
    im = param(1.0)
    fb = param(0.1)
    a = param(1.1)
    c = param(0.5)
    xi = param(0.1)
    yi = param(0.1)


@ugen(ar=True)
class FBSineL(UGen):
    """
    A linear-interpolating feedback sine with chaotic phase indexing.

    ::

        >>> fbsine_l = supriya.ugens.FBSineL.ar(
        ...     a=1.1,
        ...     c=0.5,
        ...     fb=0.1,
        ...     frequency=22050,
        ...     im=1,
        ...     xi=0.1,
        ...     yi=0.1,
        ... )
        >>> fbsine_l
        <FBSineL.ar()[0]>
    """

    frequency = param(22050)
    im = param(1.0)
    fb = param(0.1)
    a = param(1.1)
    c = param(0.5)
    xi = param(0.1)
    yi = param(0.1)


@ugen(ar=True)
class FBSineN(UGen):
    """
    A non-interpolating feedback sine with chaotic phase indexing.

    ::

        >>> fbsine_n = supriya.ugens.FBSineN.ar(
        ...     a=1.1,
        ...     c=0.5,
        ...     fb=0.1,
        ...     frequency=22050,
        ...     im=1,
        ...     xi=0.1,
        ...     yi=0.1,
        ... )
        >>> fbsine_n
        <FBSineN.ar()[0]>
    """

    frequency = param(22050)
    im = param(1.0)
    fb = param(0.1)
    a = param(1.1)
    c = param(0.5)
    xi = param(0.1)
    yi = param(0.1)


@ugen(ar=True)
class GbmanL(UGen):
    """
    A non-interpolating gingerbreadman map chaotic generator.

    ::

        >>> gbman_l = supriya.ugens.GbmanL.ar(
        ...     frequency=22050,
        ...     xi=1.2,
        ...     yi=2.1,
        ... )
        >>> gbman_l
        <GbmanL.ar()[0]>
    """

    frequency = param(22050)
    xi = param(1.2)
    yi = param(2.1)


@ugen(ar=True)
class GbmanN(UGen):
    """
    A non-interpolating gingerbreadman map chaotic generator.

    ::

        >>> gbman_n = supriya.ugens.GbmanN.ar(
        ...     frequency=22050,
        ...     xi=1.2,
        ...     yi=2.1,
        ... )
        >>> gbman_n
        <GbmanN.ar()[0]>
    """

    frequency = param(22050)
    xi = param(1.2)
    yi = param(2.1)


@ugen(ar=True)
class HenonC(UGen):
    """
    A cubic-interpolating henon map chaotic generator.

    ::

        >>> henon_c = supriya.ugens.HenonC.ar(
        ...     a=1.4,
        ...     b=0.3,
        ...     frequency=22050,
        ...     x_0=0,
        ...     x_1=0,
        ... )
        >>> henon_c
        <HenonC.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.4)
    b = param(0.3)
    x_0 = param(0)
    x_1 = param(0)


@ugen(ar=True)
class HenonL(UGen):
    """
    A linear-interpolating henon map chaotic generator.

    ::

        >>> henon_l = supriya.ugens.HenonL.ar(
        ...     a=1.4,
        ...     b=0.3,
        ...     frequency=22050,
        ...     x_0=0,
        ...     x_1=0,
        ... )
        >>> henon_l
        <HenonL.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.4)
    b = param(0.3)
    x_0 = param(0)
    x_1 = param(0)


@ugen(ar=True)
class HenonN(UGen):
    """
    A non-interpolating henon map chaotic generator.

    ::

        >>> henon_n = supriya.ugens.HenonN.ar(
        ...     a=1.4,
        ...     b=0.3,
        ...     frequency=22050,
        ...     x_0=0,
        ...     x_1=0,
        ... )
        >>> henon_n
        <HenonN.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.4)
    b = param(0.3)
    x_0 = param(0)
    x_1 = param(0)


@ugen(ar=True)
class LatoocarfianC(UGen):
    """
    A cubic-interpolating Latoocarfian chaotic generator.

    ::

        >>> latoocarfian_c = supriya.ugens.LatoocarfianC.ar(
        ...     a=1,
        ...     b=3,
        ...     c=0.5,
        ...     d=0.5,
        ...     frequency=22050,
        ...     xi=0.5,
        ...     yi=0.5,
        ... )
        >>> latoocarfian_c
        <LatoocarfianC.ar()[0]>
    """

    frequency = param(22050)
    a = param(1)
    b = param(3)
    c = param(0.5)
    d = param(0.5)
    xi = param(0.5)
    yi = param(0.5)


@ugen(ar=True)
class LatoocarfianL(UGen):
    """
    A linear-interpolating Latoocarfian chaotic generator.

    ::

        >>> latoocarfian_l = supriya.ugens.LatoocarfianL.ar(
        ...     a=1,
        ...     b=3,
        ...     c=0.5,
        ...     d=0.5,
        ...     frequency=22050,
        ...     xi=0.5,
        ...     yi=0.5,
        ... )
        >>> latoocarfian_l
        <LatoocarfianL.ar()[0]>
    """

    frequency = param(22050)
    a = param(1)
    b = param(3)
    c = param(0.5)
    d = param(0.5)
    xi = param(0.5)
    yi = param(0.5)


@ugen(ar=True)
class LatoocarfianN(UGen):
    """
    A non-interpolating Latoocarfian chaotic generator.

    ::

        >>> latoocarfian_n = supriya.ugens.LatoocarfianN.ar(
        ...     a=1,
        ...     b=3,
        ...     c=0.5,
        ...     d=0.5,
        ...     frequency=22050,
        ...     xi=0.5,
        ...     yi=0.5,
        ... )
        >>> latoocarfian_n
        <LatoocarfianN.ar()[0]>
    """

    frequency = param(22050)
    a = param(1)
    b = param(3)
    c = param(0.5)
    d = param(0.5)
    xi = param(0.5)
    yi = param(0.5)


@ugen(ar=True)
class LinCongC(UGen):
    """
    A cubic-interpolating linear congruential chaotic generator.

    ::

        >>> lin_cong_c = supriya.ugens.LinCongC.ar(
        ...     a=1.1,
        ...     c=0.13,
        ...     frequency=22050,
        ...     m=1,
        ...     xi=0,
        ... )
        >>> lin_cong_c
        <LinCongC.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.1)
    c = param(0.13)
    m = param(1)
    xi = param(0)


@ugen(ar=True)
class LinCongL(UGen):
    """
    A linear-interpolating linear congruential chaotic generator.

    ::

        >>> lin_cong_l = supriya.ugens.LinCongL.ar(
        ...     a=1.1,
        ...     c=0.13,
        ...     frequency=22050,
        ...     m=1,
        ...     xi=0,
        ... )
        >>> lin_cong_l
        <LinCongL.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.1)
    c = param(0.13)
    m = param(1)
    xi = param(0)


@ugen(ar=True)
class LinCongN(UGen):
    """
    A non-interpolating linear congruential chaotic generator.

    ::

        >>> lin_cong_n = supriya.ugens.LinCongN.ar(
        ...     a=1.1,
        ...     c=0.13,
        ...     frequency=22050,
        ...     m=1,
        ...     xi=0,
        ... )
        >>> lin_cong_n
        <LinCongN.ar()[0]>
    """

    frequency = param(22050)
    a = param(1.1)
    c = param(0.13)
    m = param(1)
    xi = param(0)


@ugen(ar=True)
class LorenzL(UGen):
    """
    A linear-interpolating Lorenz chaotic generator.

    ::

        >>> lorenz_l = supriya.ugens.LorenzL.ar(
        ...     b=2.667,
        ...     frequency=22050,
        ...     h=0.05,
        ...     r=28,
        ...     s=10,
        ...     xi=0.1,
        ...     yi=0,
        ...     zi=0,
        ... )
        >>> lorenz_l
        <LorenzL.ar()[0]>
    """

    frequency = param(22050)
    s = param(10)
    r = param(28)
    b = param(2.667)
    h = param(0.05)
    xi = param(0.1)
    yi = param(0)
    zi = param(0)


@ugen(ar=True)
class QuadC(UGen):
    """
    A cubic-interpolating general quadratic map chaotic generator.

    ::

        >>> quad_c = supriya.ugens.QuadC.ar(
        ...     a=1,
        ...     b=-1,
        ...     c=-0.75,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> quad_c
        <QuadC.ar()[0]>
    """

    frequency = param(22050)
    a = param(1)
    b = param(-1)
    c = param(-0.75)
    xi = param(0)


@ugen(ar=True)
class QuadL(UGen):
    """
    A linear-interpolating general quadratic map chaotic generator.

    ::

        >>> quad_l = supriya.ugens.QuadL.ar(
        ...     a=1,
        ...     b=-1,
        ...     c=-0.75,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> quad_l
        <QuadL.ar()[0]>
    """

    frequency = param(22050)
    a = param(1)
    b = param(-1)
    c = param(-0.75)
    xi = param(0)


@ugen(ar=True)
class QuadN(UGen):
    """
    A non-interpolating general quadratic map chaotic generator.

    ::

        >>> quad_n = supriya.ugens.QuadN.ar(
        ...     a=1,
        ...     b=-1,
        ...     c=-0.75,
        ...     frequency=22050,
        ...     xi=0,
        ... )
        >>> quad_n
        <QuadN.ar()[0]>
    """

    frequency = param(22050)
    a = param(1)
    b = param(-1)
    c = param(-0.75)
    xi = param(0)


@ugen(ar=True)
class StandardL(UGen):
    """
    A linear-interpolating standard map chaotic generator.

    ::

        >>> standard_l = supriya.ugens.StandardL.ar(
        ...     frequency=22050,
        ...     k=1,
        ...     xi=0.5,
        ...     yi=0,
        ... )
        >>> standard_l
        <StandardL.ar()[0]>
    """

    frequency = param(22050)
    k = param(1)
    xi = param(0.5)
    yi = param(0)


@ugen(ar=True)
class StandardN(UGen):
    """
    A non-interpolating standard map chaotic generator.

    ::

        >>> standard_n = supriya.ugens.StandardN.ar(
        ...     frequency=22050,
        ...     k=1,
        ...     xi=0.5,
        ...     yi=0,
        ... )
        >>> standard_n
        <StandardN.ar()[0]>
    """

    frequency = param(22050)
    k = param(1)
    xi = param(0.5)
    yi = param(0)



from ...ugens import HPF, LPF, In, Lag, ReplaceOut, SynthDef, SynthDefBuilder


def _build_synthdef() -> SynthDef:
    with SynthDefBuilder(bus=0, cutoff=0.5) as builder:
        source = In.ar(bus=builder["bus"])
        cutoff = Lag.kr(source=builder["cutoff"])
        lpf_cutoff = cutoff.clip(0.0, 0.75)
        lpf_frequency = lpf_cutoff.scale(0, 0.75, 20, 22000, True)
        hpf_cutoff = cutoff.clip(0.25, 1.0)
        hpf_frequency = hpf_cutoff.scale(0.25, 1.0, 20, 22000, True)
        source = HPF.ar(
            source=LPF.ar(source=source, frequency=lpf_frequency),
            frequency=hpf_frequency,
        )
        ReplaceOut.ar(bus=builder["bus"], source=source)
    return builder.build(name="sweep_filter")


sweep_filter = _build_synthdef()

__all__ = ["sweep_filter"]



# cython: language_level=3
# distutils: language = c++

from typing import List, Union

from .entities import Bus, BusGroup
from .shm cimport server_shared_memory_client


cdef class ServerSHM:
    """
    Server shared memory interface.

    Currently supports reading control busses only.

    .. warning::

       Not supported on Windows.

    """
    cdef server_shared_memory_client* client
    cdef unsigned int bus_count

    def __cinit__(self, unsigned int port_number, unsigned int bus_count):
        self.bus_count = bus_count
        self.client = new server_shared_memory_client(port_number)

    def __dealloc__(self) -> None:
        del self.client

    def __getitem__(self, item: Union[int, slice, Bus, BusGroup]) -> Union[float, List[float]]:
        if isinstance(item, Bus):
            item = int(item)
        elif isinstance(item, BusGroup):
            item = slice(int(item), int(item) + len(item))
        if isinstance(item, int):
            if item < 0 or item >= self.bus_count:
                raise ValueError("index out of bounds")
            return self.client.get_control_busses()[item]
        elif isinstance(item, slice):
            result = []
            for bus_index in range(*item.indices(self.bus_count)):
                result.append(self.client.get_control_busses()[bus_index])
            return result
        raise ValueError(item)

    def __setitem__(self, item: Union[int, slice, Bus, BusGroup], value: Union[float, List[float]]) -> None:
        if isinstance(item, BusGroup):
            item = slice(int(item), int(item) + len(item), 1)
        elif isinstance(item, Bus):
            item = int(item)
        if isinstance(item, int):
            if item < 0 or item >= self.bus_count:
                raise ValueError("index out of bounds")
            self.client.set_control_bus(item, float(value))
            return
        elif isinstance(item, slice):
            for value_index, bus_index in enumerate(range(*item.indices(self.bus_count))):
                value_ = float(value[value_index])
                self.client.set_control_bus(bus_index, value_)
            return
        raise ValueError(item, value)



import asyncio
import concurrent.futures
import logging
from typing import List

import pytest
from uqbar.strings import normalize

from supriya.enums import BootStatus
from supriya.osc import (
    AsyncOscProtocol,
    HealthCheck,
    OscBundle,
    OscMessage,
    ThreadedOscProtocol,
    find_free_port,
)
from supriya.osc.messages import NTP_DELTA
from supriya.scsynth import AsyncProcessProtocol, Options, SyncProcessProtocol

logger = logging.getLogger(__name__)


async def get(x):
    if asyncio.iscoroutine(x):
        return await x
    elif asyncio.isfuture(x):
        return await x
    elif isinstance(x, concurrent.futures.Future):
        return x.result()
    return x


def test_OscMessage() -> None:
    osc_message = OscMessage(
        "/foo",
        1,
        2.5,
        OscBundle(
            contents=(
                OscMessage("/bar", "baz", 3.0),
                OscMessage("/ffff", False, True, None),
            )
        ),
        ["a", "b", ["c", "d"]],
    )
    assert repr(osc_message) == normalize(
        """
    OscMessage('/foo', 1, 2.5, OscBundle(
        contents=(
            OscMessage('/bar', 'baz', 3.0),
            OscMessage('/ffff', False, True, None),
        ),
    ), ['a', 'b', ['c', 'd']])
    """
    )
    assert str(osc_message) == normalize(
        """
    size 112
       0   2f 66 6f 6f  00 00 00 00  2c 69 66 62  5b 73 73 5b   |/foo....,ifb[ss[|
      16   73 73 5d 5d  00 00 00 00  00 00 00 01  40 20 00 00   |ss]]........@ ..|
      32   00 00 00 3c  23 62 75 6e  64 6c 65 00  00 00 00 00   |...<#bundle.....|
      48   00 00 00 01  00 00 00 14  2f 62 61 72  00 00 00 00   |......../bar....|
      64   2c 73 66 00  62 61 7a 00  40 40 00 00  00 00 00 10   |,sf.baz.@@......|
      80   2f 66 66 66  66 00 00 00  2c 46 54 4e  00 00 00 00   |/ffff...,FTN....|
      96   61 00 00 00  62 00 00 00  63 00 00 00  64 00 00 00   |a...b...c...d...|
    """
    )
    datagram = osc_message.to_datagram()
    new_osc_message = OscMessage.from_datagram(datagram)
    assert osc_message == new_osc_message
    assert repr(new_osc_message) == normalize(
        """
    OscMessage('/foo', 1, 2.5, OscBundle(
        contents=(
            OscMessage('/bar', 'baz', 3.0),
            OscMessage('/ffff', False, True, None),
        ),
    ), ['a', 'b', ['c', 'd']])
    """
    )


def test_new_ntp_era() -> None:
    """
    Check for NTP timestamp overflow.
    """
    seconds = 2**32 - NTP_DELTA + 1
    datagram = OscBundle._encode_date(seconds=seconds)
    assert datagram.hex() == "0000000100000000"


@pytest.mark.parametrize(
    "osc_protocol_class, process_protocol_class",
    [
        (AsyncOscProtocol, AsyncProcessProtocol),
        (ThreadedOscProtocol, SyncProcessProtocol),
    ],
)
@pytest.mark.asyncio
async def test_OscProtocol(osc_protocol_class, process_protocol_class) -> None:
    def on_healthcheck_failed() -> None:
        healthcheck_failed.append(True)

    logger.info("START")
    healthcheck_failed: List[bool] = []
    port = find_free_port()

    logger.info("INIT PROCESS")
    process_protocol = process_protocol_class()

    logger.info("INIT PROTOCOL")
    osc_protocol = osc_protocol_class(on_panic_callback=on_healthcheck_failed)
    assert osc_protocol.status == BootStatus.OFFLINE

    try:
        logger.info("BOOT PROCESS")
        await get(process_protocol.boot(Options(port=port)))
        await get(process_protocol.boot_future)
        assert process_protocol.status == BootStatus.ONLINE

        logger.info("CONNECT PROTOCOL")
        await get(
            osc_protocol.connect(
                "127.0.0.1",
                port,
                healthcheck=HealthCheck(
                    request_pattern=["/status"],
                    response_pattern=["/status.reply"],
                    max_attempts=3,
                ),
            )
        )
        assert osc_protocol.status == BootStatus.BOOTING

        logger.info("AWAIT CONNECTION")
        await get(osc_protocol.boot_future)
        assert osc_protocol.status == BootStatus.ONLINE

        logger.info("QUIT PROCESS")
        await get(process_protocol.quit())

        logger.info("AWAIT DISCONNECTION")
        await get(osc_protocol.exit_future)
        assert osc_protocol.status == BootStatus.OFFLINE

    finally:
        await get(process_protocol.quit())



import asyncio
import logging
import queue
import traceback
from typing import Awaitable, Optional, Tuple

from .bases import BaseClock
from .ephemera import (
    Action,
    CallbackEvent,
    ChangeEvent,
    ClockContext,
    Command,
    Moment,
    TimeUnit,
)

logger = logging.getLogger(__name__)


class AsyncClock(BaseClock):
    def __init__(self) -> None:
        BaseClock.__init__(self)
        self._task: Optional[Awaitable[None]] = None
        self._slop = 1.0
        try:
            self._event = asyncio.Event()
        except RuntimeError:
            pass

    ### SCHEDULING METHODS ###

    def _enqueue_command(self, command: Command) -> None:
        super()._enqueue_command(command)
        self._event.set()

    async def _perform_callback_event_async(
        self, event: CallbackEvent, current_moment: Moment, desired_moment: Moment
    ) -> None:
        logger.debug(
            f"[{self.name}] ... ... Performing {event.procedure} at "
            f"{desired_moment.seconds - self._state.initial_seconds}:s / "
            f"{desired_moment.offset}:o"
        )
        context = ClockContext(current_moment, desired_moment, event)
        args = event.args or ()
        kwargs = event.kwargs or {}
        try:
            result = event.procedure(context, *args, **kwargs)
            if asyncio.iscoroutine(result):
                result = await result
        except Exception:
            traceback.print_exc()
            return
        if isinstance(result, float) or result is None:
            delta, time_unit = result, TimeUnit.BEATS
        else:
            delta, time_unit = result
        self._process_callback_event_result(desired_moment, event, delta, time_unit)

    async def _perform_events_async(self, current_moment: Moment) -> Moment:
        logger.debug(
            f"[{self.name}] ... Ready to perform at "
            f"{current_moment.seconds - self._state.initial_seconds}:s / "
            f"{current_moment.offset}:o"
        )
        while self._is_running and self._event_queue.qsize():
            (
                event,
                desired_moment,
                should_continue,
                should_break,
            ) = self._process_perform_event_loop(current_moment)
            if should_continue:
                continue
            elif should_break:
                break
            if event is None or desired_moment is None:
                raise ValueError(event, desired_moment)
            if isinstance(event, ChangeEvent):
                current_moment, should_continue = self._perform_change_event(
                    event, current_moment, desired_moment
                )
                if not should_continue:
                    break
            elif isinstance(event, CallbackEvent):
                await self._perform_callback_event_async(
                    event, current_moment, desired_moment
                )
                self._process_command_deque()
            else:
                raise ValueError(event)
        return current_moment

    async def _run_async(self, offline: bool = False) -> None:
        logger.debug(f"[{self.name}] Coroutine start")
        self._process_command_deque(first_run=True)
        while self._is_running:
            logger.debug(f"[{self.name}] Loop start")
            if not await self._wait_for_queue_async():
                return
            try:
                if (current_moment := await self._wait_for_moment_async()) is None:
                    return
            except queue.Empty:
                continue
            current_moment = await self._perform_events_async(current_moment)
            self._state = self._state._replace(
                previous_seconds=current_moment.seconds,
                previous_offset=current_moment.offset,
            )
        logger.debug(f"[{self.name}] Coroutine terminating")
        self._stop()

    async def _wait_for_event_async(self, sleep_time: float) -> None:
        try:
            await asyncio.wait_for(self._event.wait(), sleep_time)
        except (asyncio.TimeoutError, RuntimeError):
            pass

    async def _wait_for_moment_async(self, offline: bool = False) -> Optional[Moment]:
        current_time = self.get_current_time()
        next_time = self._event_queue.peek().seconds
        logger.debug(
            f"[{self.name}] ... Waiting for next moment at {next_time} from {current_time}"
        )
        while current_time < next_time:
            if not offline:
                await self._wait_for_event_async(next_time - current_time)
            if not self._is_running:
                return None
            self._process_command_deque()
            next_time = self._event_queue.peek().seconds
            current_time = self.get_current_time()
            self._event.clear()
        return self._seconds_to_moment(current_time)

    async def _wait_for_queue_async(self, offline: bool = False) -> bool:
        logger.debug(f"[{self.name}] ... Waiting for events")
        self._process_command_deque()
        self._event.clear()
        while not self._event_queue.qsize():
            if not offline:
                await self._event.wait()
            if not self._is_running:
                return False
            self._process_command_deque()
            self._event.clear()
        return True

    ### PUBLIC METHODS ###

    def cancel(self, event_id: int) -> Optional[Action]:
        event = super().cancel(event_id)
        self._event.set()
        return event

    async def start(
        self,
        initial_time: Optional[float] = None,
        initial_offset: float = 0.0,
        initial_measure: int = 1,
        beats_per_minute: Optional[float] = None,
        time_signature: Optional[Tuple[int, int]] = None,
    ) -> None:
        self._start(
            initial_time=initial_time,
            initial_offset=initial_offset,
            initial_measure=initial_measure,
            beats_per_minute=beats_per_minute,
            time_signature=time_signature,
        )
        loop = asyncio.get_running_loop()
        self._event = asyncio.Event()
        self._task = loop.create_task(self._run_async())

    async def stop(self) -> None:
        if self._stop():
            self._event.set()
            if self._task is not None:
                await self._task



import asyncio
import concurrent.futures
import logging
import random
import sys
from pathlib import Path

import pytest
import pytest_asyncio

from supriya import assets, default
from supriya.contexts.errors import MomentClosed
from supriya.contexts.realtime import AsyncServer, Server
from supriya.contexts.responses import BufferInfo
from supriya.osc import OscBundle, OscMessage


async def get(x):
    if asyncio.iscoroutine(x):
        return await x
    return x


@pytest.fixture
def audio_paths():
    return sorted((Path(assets.__path__[0]) / "audio" / "birds").glob("*.wav"))


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest_asyncio.fixture(autouse=True, params=[AsyncServer, Server])
async def context(request):
    context = request.param()
    await get(context.boot())
    context.add_synthdefs(default)
    await get(context.sync())
    yield context


@pytest.mark.asyncio
async def test_Buffer_allocated(context):
    # buffers can be allocated
    buffer_a = context.add_buffer(channel_count=1, frame_count=512)
    buffer_b = context.add_buffer(channel_count=1, frame_count=512)
    assert not buffer_a.allocated  # waiting for the /done /b_alloc response
    assert not buffer_b.allocated  # waiting for the /done /b_alloc response
    await get(context.sync())
    assert buffer_a.allocated
    assert buffer_b.allocated
    # freed buffers are unallocated
    buffer_a.free()
    assert buffer_a.allocated  # waiting for the /done /b_free response
    assert buffer_b.allocated
    await get(context.sync())
    assert not buffer_a.allocated  # waiting for the /done /b_free response
    assert buffer_b.allocated
    # nothing is allocated on an unbooted context
    await get(context.quit())
    assert not buffer_a.allocated
    assert not buffer_b.allocated
    # allocation state is stored in the context, not the node
    await get(context.boot())
    assert not buffer_a.allocated
    assert not buffer_b.allocated


@pytest.mark.asyncio
async def test_add_buffer(audio_paths, context):
    with context.osc_protocol.capture() as transcript:
        # neither frame count nor file path provided
        with pytest.raises(ValueError):
            context.add_buffer()
        # both channel count and file path provided
        with pytest.raises(ValueError):
            context.add_buffer(file_path=audio_paths[0], channel_count=1)
        # both channel count and channel indices provided
        with pytest.raises(ValueError):
            context.add_buffer(
                file_path=audio_paths[0], channel_count=1, channel_indices=[0]
            )
        # /b_alloc
        buffer_a = context.add_buffer(channel_count=1, frame_count=23)
        # /b_allocRead
        buffer_b = context.add_buffer(frame_count=19, file_path=audio_paths[0])
        # /b_allocReadChannel
        buffer_c = context.add_buffer(
            frame_count=17, file_path=audio_paths[1], channel_indices=[1, 0]
        )
        # completion without moment errors
        with pytest.raises(MomentClosed):
            buffer_d = context.add_buffer(channel_count=3, frame_count=31)
            with buffer_d:
                ...
        # completion without moment via on_completion lambda succeeds
        buffer_e = context.add_buffer(
            channel_count=1, frame_count=47, on_completion=lambda ctx: ctx.add_group()
        )
        # completion inside moment succeeds
        with context.at(1.23):
            buffer_f = context.add_buffer(channel_count=3, frame_count=31)
            with buffer_f:
                context.add_group()
    buffers = [buffer_a, buffer_b, buffer_c, buffer_d, buffer_e, buffer_f]
    assert all(buffer.context is context for buffer in buffers)
    assert [buffer.id_ for buffer in buffers] == [0, 1, 2, 3, 4, 5]
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_alloc", 0, 23, 1),
        OscMessage("/b_allocRead", 1, str(audio_paths[0]), 0, 19),
        OscMessage("/b_allocReadChannel", 2, str(audio_paths[1]), 0, 17, 1, 0),
        OscMessage("/b_alloc", 3, 31, 3),
        OscMessage("/b_alloc", 4, 47, 1, OscMessage("/g_new", 1000, 0, 1)),
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 5, 31, 3, OscMessage("/g_new", 1001, 0, 1)),
            ),
            timestamp=1.23 + context.latency,
        ),
    ]


@pytest.mark.asyncio
async def test_add_buffer_group(context):
    # neither channel count nor frame count provided
    with pytest.raises(ValueError):
        context.add_buffer_group()
    # count less than 1
    with pytest.raises(ValueError):
        context.add_buffer_group(channel_count=1, count=0, frame_count=512)
    # /b_alloc
    with context.osc_protocol.capture() as transcript:
        buffer_group = context.add_buffer_group(
            channel_count=1, count=5, frame_count=23
        )
    assert len(buffer_group) == 5
    assert all(buffer.context is context for buffer in buffer_group)
    assert [buffer.id_ for buffer in buffer_group] == [0, 1, 2, 3, 4]
    assert transcript.filtered(received=False, status=False) == [
        OscBundle(
            contents=(
                OscMessage("/b_alloc", 0, 23, 1),
                OscMessage("/b_alloc", 1, 23, 1),
                OscMessage("/b_alloc", 2, 23, 1),
                OscMessage("/b_alloc", 3, 23, 1),
                OscMessage("/b_alloc", 4, 23, 1),
            )
        )
    ]


@pytest.mark.asyncio
async def test_close_buffer(context):
    buffer_a = context.add_buffer(channel_count=1, frame_count=23)
    buffer_b = context.add_buffer(channel_count=1, frame_count=23)
    buffer_c = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        # completion without moment errors, but initial request succeeds
        with pytest.raises(MomentClosed):
            with buffer_a.close():
                ...
        # completion without moment via on_completion lambda succeeds
        buffer_b.close(on_completion=lambda ctx: buffer_b.free())
        # completion inside moment succeeds
        with context.at():
            with buffer_c.close():
                buffer_c.free()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_close", 0),
        OscMessage("/b_close", 1, OscMessage("/b_free", 1)),
        OscMessage("/b_close", 2, OscMessage("/b_free", 2)),
    ]


@pytest.mark.asyncio
async def test_copy_buffer(context):
    buffer_a = context.add_buffer(channel_count=1, frame_count=23)
    buffer_b = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        buffer_a.copy(
            target_buffer=buffer_b,
            starting_frame=5,
            target_starting_frame=7,
            frame_count=3,
        )
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_gen", 1, "copy", 7, 0, 5, 3)
    ]


@pytest.mark.asyncio
async def test_fill_buffer(context):
    buffer_a = context.add_buffer(channel_count=1, frame_count=23)
    buffer_b = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        # one-off
        buffer_a.fill(2, 3, 0.5)
        # with moment, including grouping
        with context.at():
            buffer_a.fill(2, 3, 0.6)
            buffer_b.fill(3, 4, 0.5)
            buffer_a.fill(4, 5, 0.4)
            buffer_a.fill(5, 6, 0.3)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_fill", 0, 2, 3, 0.5),
        OscBundle(
            contents=(
                OscMessage("/b_fill", 0, 2, 3, 0.6, 4, 5, 0.4, 5, 6, 0.3),
                OscMessage("/b_fill", 1, 3, 4, 0.5),
            )
        ),
    ]


@pytest.mark.asyncio
async def test_free_buffer(context):
    buffer_a = context.add_buffer(channel_count=1, frame_count=23)
    buffer_b = context.add_buffer(channel_count=1, frame_count=23)
    buffer_c = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        # completion without moment errors, but initial request succeeds
        with pytest.raises(MomentClosed):
            with buffer_a.free():
                ...
        # completion without moment via on_completion lambda succeeds
        buffer_b.free(on_completion=lambda ctx: ctx.add_group())
        # completion inside moment succeeds
        with context.at():
            with buffer_c.free():
                context.add_group()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_free", 0),
        OscMessage("/b_free", 1, OscMessage("/g_new", 1000, 0, 1)),
        OscMessage("/b_free", 2, OscMessage("/g_new", 1001, 0, 1)),
    ]


def test_generate_buffer(context):
    buffer = context.add_buffer(channel_count=1, frame_count=1024)
    with context.osc_protocol.capture() as transcript:
        with context.at(0):
            buffer.generate(command_name="sine1", amplitudes=[1, 2, 3])
            buffer.generate(
                command_name="sine2", amplitudes=[1, 2, 3], frequencies=[4, 5, 6]
            )
            buffer.generate(
                command_name="sine3",
                amplitudes=[1, 2, 3],
                frequencies=[4, 5, 6],
                phases=[0.25, 0.0, 0.5],
            )
            buffer.generate(command_name="cheby", amplitudes=[1, 2, 3])
            buffer.generate(
                command_name="sine1", amplitudes=[1, 2, 3], as_wavetable=True
            )
            buffer.generate(
                command_name="sine1", amplitudes=[1, 2, 3], should_clear_first=True
            )
            buffer.generate(
                command_name="sine1",
                amplitudes=[1, 2, 3],
                should_clear_first=True,
                should_normalize=True,
            )
    assert transcript.filtered(received=False, status=False) == [
        OscBundle(
            contents=(
                OscMessage("/b_gen", 0, "sine1", 0, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine2", 0, 4.0, 1.0, 5.0, 2.0, 6.0, 3.0),
                OscMessage(
                    "/b_gen",
                    0,
                    "sine3",
                    0,
                    4.0,
                    1.0,
                    0.25,
                    5.0,
                    2.0,
                    0.0,
                    6.0,
                    3.0,
                    0.5,
                ),
                OscMessage("/b_gen", 0, "cheby", 0, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine1", 2, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine1", 4, 1.0, 2.0, 3.0),
                OscMessage("/b_gen", 0, "sine1", 5, 1.0, 2.0, 3.0),
            ),
            timestamp=0.0 + context.latency,
        ),
    ]


@pytest.mark.asyncio
async def test_get_buffer(context):
    buffer = context.add_buffer(channel_count=1, frame_count=512)
    exception_classes = (concurrent.futures.TimeoutError,)
    if sys.version_info >= (3, 8):
        exception_classes += (asyncio.exceptions.TimeoutError,)
    with pytest.raises(exception_classes):
        await get(buffer.get(1))
    await get(context.sync())
    assert await get(buffer.get(1, 2, 3)) == {1: 0.0, 2: 0.0, 3: 0.0}
    # unsync
    with context.osc_protocol.capture() as transcript:
        assert await get(buffer.get(1, sync=False)) is None
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_get", 0, 1)
    ]


@pytest.mark.asyncio
async def test_get_buffer_range(context):
    buffer = context.add_buffer(channel_count=1, frame_count=512)
    exception_classes = (concurrent.futures.TimeoutError,)
    if sys.version_info >= (3, 8):
        exception_classes += (asyncio.exceptions.TimeoutError,)
    with pytest.raises(exception_classes):
        await get(context.get_buffer_range(100, 1, 3))  # no such buffer exists
    await get(context.sync())
    assert await get(buffer.get_range(1, 3)) == (0.0, 0.0, 0.0)
    # unsync
    with context.osc_protocol.capture() as transcript:
        assert await get(buffer.get_range(1, 3, sync=False)) is None
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_getn", 0, 1, 3)
    ]


@pytest.mark.asyncio
async def test_normalize_buffer(context):
    buffer = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        buffer.normalize(0.5)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_gen", 0, "normalize", 0.5)
    ]


@pytest.mark.asyncio
async def test_query_buffer(context):
    frame_count = random.randint(128, 256)
    channel_count = random.randint(1, 8)
    buffer = context.add_buffer(channel_count=channel_count, frame_count=frame_count)
    await asyncio.sleep(0.1)
    assert await get(buffer.query()) in [
        BufferInfo(
            items=[
                BufferInfo.Item(
                    buffer_id=0,
                    frame_count=frame_count,
                    channel_count=channel_count,
                    sample_rate=sample_rate,
                )
            ]
        )
        for sample_rate in [48000, 44100]
    ]
    # unsync
    with context.osc_protocol.capture() as transcript:
        assert await get(buffer.query(sync=False)) is None
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_query", 0)
    ]


@pytest.mark.asyncio
async def test_read_buffer(audio_paths, context):
    buffer_a = context.add_buffer(channel_count=1, frame_count=23)
    buffer_b = context.add_buffer(channel_count=1, frame_count=23)
    buffer_c = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        # completion without moment errors, but initial request succeeds
        with pytest.raises(MomentClosed):
            with buffer_a.read(file_path=audio_paths[0]):
                ...
        # completion without moment via on_completion lambda succeeds
        buffer_b.read(
            file_path=audio_paths[1], on_completion=lambda ctx: ctx.add_group()
        )
        # completion inside moment succeeds
        with context.at():
            with buffer_c.read(file_path=audio_paths[2]):
                context.add_group()
        # parameters
        buffer_a.read(
            buffer_starting_frame=5,
            channel_indices=[0],
            file_path=audio_paths[3],
            frame_count=16,
            leave_open=True,
            starting_frame=3,
        )
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_read", 0, str(audio_paths[0]), 0, -1, 0, 0),
        OscMessage(
            "/b_read",
            1,
            str(audio_paths[1]),
            0,
            -1,
            0,
            0,
            OscMessage("/g_new", 1000, 0, 1),
        ),
        OscMessage(
            "/b_read",
            2,
            str(audio_paths[2]),
            0,
            -1,
            0,
            0,
            OscMessage("/g_new", 1001, 0, 1),
        ),
        OscMessage("/b_readChannel", 0, str(audio_paths[3]), 3, 16, 5, 1, 0),
    ]


@pytest.mark.asyncio
async def test_set_buffer(context):
    buffer = context.add_buffer(channel_count=1, frame_count=32)
    with context.osc_protocol.capture() as transcript:
        buffer.set(2, 0.5)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_set", 0, 2, 0.5)
    ]


@pytest.mark.asyncio
async def test_set_buffer_range(context):
    buffer = context.add_buffer(channel_count=1, frame_count=32)
    with context.osc_protocol.capture() as transcript:
        buffer.set_range(4, (0.5, 0.75, 0.25))
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_setn", 0, 4, 3, 0.5, 0.75, 0.25)
    ]


@pytest.mark.asyncio
async def test_write_buffer(context, tmp_path):
    buffer_a = context.add_buffer(channel_count=1, frame_count=23)
    buffer_b = context.add_buffer(channel_count=1, frame_count=23)
    buffer_c = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        # completion without moment errors, but initial request succeeds
        with pytest.raises(MomentClosed):
            with buffer_a.write(file_path=tmp_path / "foo-1.aiff"):
                ...
        # completion without moment via on_completion lambda succeeds
        buffer_b.write(
            file_path=tmp_path / "foo-2.aiff", on_completion=lambda ctx: ctx.add_group()
        )
        # completion inside moment succeeds
        with context.at():
            with buffer_c.write(file_path=tmp_path / "foo-3.aiff"):
                context.add_group()
        # parameters
        buffer_a.write(
            file_path=tmp_path / "foo-4.wav",
            frame_count=16,
            header_format="WAV",
            leave_open=True,
            sample_format="FLOAT",
            starting_frame=5,
        )
    assert transcript.filtered(received=False, status=False) == [
        OscMessage(
            "/b_write", 0, str(tmp_path / "foo-1.aiff"), "aiff", "int24", -1, 0, 0
        ),
        OscMessage(
            "/b_write",
            1,
            str(tmp_path / "foo-2.aiff"),
            "aiff",
            "int24",
            -1,
            0,
            0,
            OscMessage("/g_new", 1000, 0, 1),
        ),
        OscMessage(
            "/b_write",
            2,
            str(tmp_path / "foo-3.aiff"),
            "aiff",
            "int24",
            -1,
            0,
            0,
            OscMessage("/g_new", 1001, 0, 1),
        ),
        OscMessage(
            "/b_write", 0, str(tmp_path / "foo-4.wav"), "wav", "float", 16, 5, 1
        ),
    ]


@pytest.mark.asyncio
async def test_zero_buffer(context):
    buffer_a = context.add_buffer(channel_count=1, frame_count=23)
    buffer_b = context.add_buffer(channel_count=1, frame_count=23)
    buffer_c = context.add_buffer(channel_count=1, frame_count=23)
    with context.osc_protocol.capture() as transcript:
        # completion without moment errors, but initial request succeeds
        with pytest.raises(MomentClosed):
            with buffer_a.zero():
                ...
        # completion without moment via on_completion lambda succeeds
        buffer_b.zero(on_completion=lambda ctx: ctx.add_group())
        # completion inside moment succeeds
        with context.at():
            with buffer_c.zero():
                context.add_group()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/b_zero", 0),
        OscMessage("/b_zero", 1, OscMessage("/g_new", 1000, 0, 1)),
        OscMessage("/b_zero", 2, OscMessage("/g_new", 1001, 0, 1)),
    ]



from typing import Optional, Sequence

from supriya import utils

from ..enums import CalculationRate
from .core import UGen, UGenOperable, UGenRecursiveInput, UGenScalarInput, param, ugen


@ugen(ar=True, kr=True)
class Blip(UGen):
    """
    A band limited impulse generator.

    ::

        >>> blip = supriya.ugens.Blip.ar(
        ...     frequency=440,
        ...     harmonic_count=200,
        ... )
        >>> blip
        <Blip.ar()[0]>
    """

    frequency = param(440.0)
    harmonic_count = param(200.0)


@ugen(ar=True, kr=True)
class FSinOsc(UGen):
    """
    Very fast sine wave generator (2 PowerPC instructions per output sample!) implemented using a ringing filter.

    ::

        >>> fsin_osc = supriya.ugens.FSinOsc.ar(
        ...     frequency=440,
        ...     initial_phase=0,
        ... )
        >>> fsin_osc
        <FSinOsc.ar()[0]>
    """

    frequency = param(440.0)
    initial_phase = param(0.0)


@ugen(ar=True)
class Klank(UGen):
    """
    A bank of resonators.

    ::

        >>> klank = supriya.ugens.Klank.ar(
        ...     amplitudes=None,
        ...     decay_scale=1,
        ...     decay_times=[1, 1, 1, 1],
        ...     frequencies=[200, 671, 1153, 1723],
        ...     frequency_offset=0,
        ...     frequency_scale=1,
        ...     source=supriya.ugens.BrownNoise.ar() * 0.001,
        ... )
        >>> klank
        <Klank.ar()[0]>
    """

    source = param()
    frequency_scale = param(1)
    frequency_offset = param(0)
    decay_scale = param(1)
    specifications = param(unexpanded=True)

    @classmethod
    def ar(
        cls,
        *,
        amplitudes: Optional[Sequence[UGenScalarInput]] = None,
        decay_scale: UGenRecursiveInput = 1,
        decay_times: Optional[Sequence[UGenScalarInput]] = None,
        frequencies: Sequence[UGenScalarInput],
        frequency_offset: UGenRecursiveInput = 0,
        frequency_scale: UGenRecursiveInput = 1,
        source: UGenRecursiveInput,
    ) -> UGenOperable:
        if not frequencies:
            raise ValueError(frequencies)
        if not amplitudes:
            amplitudes = [1.0] * len(frequencies)
        if not decay_times:
            decay_times = [1.0] * len(frequencies)
        specifications: Sequence[UGenScalarInput] = list(
            utils.flatten(utils.zip_cycled(frequencies, amplitudes, decay_times))
        )
        return cls._new_expanded(
            calculation_rate=CalculationRate.AUDIO,
            decay_scale=decay_scale,
            frequency_offset=frequency_offset,
            frequency_scale=frequency_scale,
            source=source,
            specifications=specifications,
        )


@ugen(ar=True, kr=True)
class Pulse(UGen):
    """
    Band limited pulse wave generator with pulse width modulation.

    ::

        >>> pulse = supriya.ugens.Pulse.ar(
        ...     frequency=440,
        ...     width=0.5,
        ... )
        >>> pulse
        <Pulse.ar()[0]>
    """

    frequency = param(440.0)
    width = param(0.5)


@ugen(ar=True, kr=True, is_pure=True)
class Saw(UGen):
    """
    A band-limited sawtooth oscillator unit generator.

    ::

        >>> supriya.ugens.Saw.ar()
        <Saw.ar()[0]>
    """

    frequency = param(440.0)
